import arrow
# import json

class DisplayPage:
    """ base class for each display """
    def __init__(self, dba, matrix=None):
        self.dba = dba
        if matrix is not None:
            self.my_matrix = matrix
            self.my_canvas = matrix.CreateFrameCanvas()
        self.values = None
        self.nextUpdate = None
        self.offsetCounter = 1
        self.is_paused = False
        self.data_dirty = True # only redraw on new data

    def check(self,now):
        """ If it's time for the next update, call update() """
        if self.nextUpdate is None or self.nextUpdate < now:
            print(f"Updating {self.type} at {now.format('MM/DD/YYYY h:mm:ss A ZZZ')}")
            self.update()

    def update(self): # implements a back off mechanism for stale messages
        """ Reads a record of self.type, 
            if it returns a new record (Happy)
            else backs off and tries again
            !!! Back off doesn't work right ;-/ !!!
        """
        self.values = self.dba.read(self.type)
        if self.values is not None: # got good data
            self.data_dirty = True # we have new data
            self.values['valid'] = self.fix_edt(self.values['valid'])
            next = arrow.get(self.values['valid'],'MM/DD/YYYY h:mm:ss A Z') # next holds the new validity date
            if self.nextUpdate == next.shift(minutes=+self.offsetCounter): # we got a stale message
                self.offsetCounter += 1
                self.nextUpdate = next.shift(minutes=+self.offsetCounter) #check again in 1 minute
            else: # we got a new message/validty time
                self.nextUpdate = next # update the next update to the new time
                self.offsetCounter = 1

    def display(self): # override this
        """ virtual function """
        # if self.data_dirty:
        #     draw new stuff
        # else:
        #     just throw up the old self.icon
        pass
    
    def justify(self, text, font, x=0, y=0, anchor='TL'):
        """ routine to logically position text """
        # width, height = font.getsize(text) # getsize went away at some version of Pillow
        # left, top, right, bottom = font.getbbox(text) # kinda goofy, left and top = 0??? 
        # width = right - left
        # height = bottom - top
        left, top, width, height = font.getbbox(text) 
        # Calculate the 'Y'
        if anchor[0] == 'T':   # Top
            yout = y   # previous bug?!?
        elif anchor[0] == 'M': # Middle (vertical)
            yout = y - height/2
        else:                  # Bottom
            yout = y + height
        # Calculate the 'X'    
        if anchor[1] == 'R':   # Right
            xout = x - width
        elif anchor[1] == 'C': # Center (horizontal)
            xout = x - width/2
        else:                  # Left
            xout = x
        return (xout, yout)

    def fix_edt(self,timestr):
        """" Fix non-uniqueness of the 'EDT' timezone abbreviation :-/ """
        if 'EDT' not in timestr:
            return timestr
        else:
            # replace EDT with US/Eastern
            # this is a hack to fix the non-uniqueness of the EDT timezone abbreviation
            # which is used in the US and Canada, but not in Europe
            # see
            return timestr.replace('EDT','US/Eastern')
    
    def stringToDateTime(self,isostring):  # formerly 'stringToTime'
        #accepts an ISO date/time string and returns a time object based on the hour and minute (sets seconds to 0)
        return arrow.get(isostring).floor('second')

    def string_to_tuple(self, str):
        tuplestr = str.replace('(','').replace(')','').split(',')
        return (int(tuplestr[0]), int(tuplestr[1]), int(tuplestr[2]))

    def hexToTuple(self,h):
        return (int(h[0:2], 16),int(h[2:4], 16),int(h[4:6], 16))

    def suffix(self,num):
        if num == 1:
            return 'st'
        elif num == 2:
            return 'nd'
        elif num ==3:
            return 'rd'
        else:
            return 'th'

    def humanlat(self, latstr):
        lat = float(latstr)
        if lat > 0:
            ns = 'N'
        elif lat < 0:
            ns = 'S'
        else:
            ns = ' '
        return ' {:>7.3f} {}'.format(lat, ns)

    def humanlon(self, lonstr):
        lon = float(lonstr)
        if lon > 0:
            ew = 'E'
        elif lon < 0:
            ew = 'W'
            lon = -1 * lon
        else:
            ew = ' '
        return '{:>8.3f} {}'.format(lon, ew)

    def shortentime(self, timestr):
        return arrow.get(timestr,'M/D/YYYY h:mm:ss A').format('M/D/YYYY h:mmA')

    def uptime_color(self, now, then):
        # print(int(float(tnow.format('X')) - float(client_start_time.format('X'))))
        time = int(float(now.format('X')) - float(then.format('X')))
        if   time < 60 * 15:      # 15 minutes
            return "Orange"
        elif time < 60 * 30:      # 30 minutes
            return "Yellow"
        elif time < 60 * 60 * 1:  #  1 hour
            return "Green"
        elif time < 60 * 60 * 6:  #  6 hours
            return "Turquoise"
        else:
            return "Blue"


""" ... """
import time
import json

import arrow
import socketserver                         # for liveness probe (MyTCPHandler(), TCPServer)
import requests   
from   requests.exceptions import HTTPError #, ConnectionError
from   requests.adapters   import HTTPAdapter, Retry
import redis

from   datasourcelib       import Database          # wrapper for postgres/cockroach/sqlite/mongodb
from   secretmanager       import SecretManager     # for reading secrets from Vault, K8s, or environment variables

class ServerPage:
    """ ... """
    def __init__(self, prod: bool, period: int, secretcfg: dict, secretdef: dict):
        self.prod = prod
        self.type = None
        self.rsess = requests.session()
        retries = Retry(total=5,
                        backoff_factor=0.5,
                        status_forcelist=[500, 502, 503, 504])
        self.rsess.mount('https://', HTTPAdapter(max_retries=retries))
        self.secrets = self.read_secrets(secretcfg, secretdef)
        self.timezone = self.secrets.get('timezone')
        self.dba = Database(self.secrets['db_type'], self.secrets['db_params'])
        if self.secrets['rhost']:
            self.r = self.connect_redis()
        self.update_period = period
        self.last_update = 0
        self.output = False

    def read_secrets(self, secretcfg: dict, secretdef: dict):
        """ Read secrets from a file, Vault, Kubernetes, or environment variables. """
        sm = SecretManager(secretcfg, log_level='INFO')
        if secretcfg.get("SOURCE") == "FILE":
            # Read secrets from a file
            secrets = sm.read_secrets(secretdef.get("file_name"), secretdef.get("file_type"))
        elif secretcfg.get("SOURCE") == "ENVIRONMENT":
            # Read secrets from environment variables
            secrets = sm.read_secrets(secretdef.get("environment_definition"), secretdef.get("env_file"), 
                                      secretdef.get("definition_type"))
        elif secretcfg.get("SOURCE") == "KUBERNETES":
            # Read secrets from Kubernetes
            secrets = sm.read_secrets(secretdef.get("secret_name"), secretdef.get("namespace"),
                                  secretdef.get("read_type"))
        elif secretcfg.get("SOURCE") == "KUBEVAULT":
            
               secrets = sm.read_secrets(secretdef.get("secret_name"), secretdef.get("namespace"), 
                                  secretdef.get("read_type"), secretdef.get("secret_key"), 
                                  secretdef.get("transit_key"))
        else:
            print(f"Unknown secret source: {secretcfg.get('SOURCE')}")
            return None
        del sm  # Clean up the SecretManager instance
        return secrets

    def connect_redis(self):
        if self.secrets['rpass'] == 'None':
            r = redis.Redis(host=self.secrets['rhost'], port=6379, db=0, decode_responses=True)
        else:
            r = redis.Redis(host=self.secrets['rhost'], port=6379, db=0, decode_responses=True, 
                                 password=self.secrets['rpass'])
        return r
    
    def update(self): # really must be overridden...
        """ ... """
        print(f"{type(self).__name__} updated.")

    def check(self, now: float):
        """ ... """
        if self.last_update == 0 or now - self.last_update > self.update_period:
            self.last_update = now
            print(f'Updating {type(self).__name__}...')
            self.update()
            if hasattr(self, "r") and self.r is not None:
                self.r.publish('update', self.type)
            print(f"{arrow.now().to(self.timezone).format('MM/DD/YYYY h:mm:ss A ZZZ')}: " \
                  f"{type(self).__name__} updated.")

    def run(self):
        """ ... """
        # Write Startup record to database
        tnow = arrow.now()
        data = {}
        data['type']   = f'{self.type}-Server'
        data['updated'] = tnow.to(self.timezone).format('MM/DD/YYYY h:mm A Z')
        data['valid']   = tnow.to(self.timezone).format('MM/DD/YYYY h:mm:ss A Z')
        data['values'] = {}
        self.dba.write(data)

        if self.prod:
            HOST = '0.0.0.0'
            PORT = 10255

            class MyTCPHandler(socketserver.BaseRequestHandler):
                # Socket handler for liveness probe
                def handle(self):
                    self.data = self.request.recv(1024).strip()
                    print('Health Check...')
                    self.request.sendall(self.data.upper())

            with socketserver.TCPServer((HOST, PORT), MyTCPHandler) as server:
                server.timeout = 0.1
                while True:
                    now = time.monotonic()
                    self.check(now)
                    server.handle_request()
                    time.sleep(0.9)
        else:
            while True:
                    now = time.monotonic()
                    self.check(now)
                    # server.handle_request()
                    time.sleep(0.95)


    def fetch(self, url: str, name: str, now: str, auth: str=None, headers: str=None): 
        """ ... """
        with self.rsess as sess:
            try:
                if auth is None and headers is None:
                    response = sess.get(url,timeout=(5,15))
                elif headers is None:
                    response = sess.get(url,timeout=(5,15),auth=auth)
                else:
                    response = sess.get(url,timeout=(5,15),headers=headers)
                response.raise_for_status()
            except HTTPError as http_err:
                print(f'({name}) HTTP error occurred: {http_err} @ {now}')
                return None
            except ConnectionError as http_err:
                print(f'({name}) HTTP error occurred: {http_err} @ {now}')
                return None
            # except Exception as err:
            #     print(f'({name}) Other error occurred: {err} @ {now}')
            #     return None
            else:
                try:
                    if self.output:
                        print(response)
                        print(f'apparent_encoding: [{response.apparent_encoding}]')
                        print(f'content: [{response.content}]')
                        print(f'   text: [{response.text}]')
                    values = response.json()
                except json.decoder.JSONDecodeError:
                    raise
                else:
                    return values # response.json()

    def fetch_raw(self, url: str, name: str, now: str):
        """ required for Garmin server which expects an XML response instead of a JSON one """
        with self.rsess as sess:
            try:
                response = sess.get(url,timeout=(5,15))
                response.raise_for_status()
            except HTTPError as http_err:
                print(f'({name}) HTTP error occurred: {http_err} @ {now}')
                return None
            except ConnectionError as http_err:
                print(f'({name}) HTTP error occurred: {http_err} @ {now}')
                return None
            # except Exception as err:
            #     print(f'({name}) Other error occurred: {err} @ {now}')
            #     return None
            return response

    def now_str(self, now: arrow, secs: bool):
        """ ... """
        if secs:
            return now.format('MM/DD/YYYY h:mm:ss A Z')

        return now.format('MM/DD/YYYY h:mm A Z')
    
